# -*- coding: utf-8 -*-
"""Utilities for generating OpenAPI spec (fka Swagger) entities from
marshmallow :class:`Schemas <marshmallow.Schema>` and :class:`Fields <marshmallow.fields.Field>`.

OpenAPI 2.0 spec: https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md
"""
from __future__ import absolute_import, unicode_literals
import operator
import warnings
import functools
from collections import OrderedDict
import copy

import marshmallow
from marshmallow.utils import is_collection
from marshmallow.compat import iteritems
from marshmallow.orderedset import OrderedSet

from apispec.lazy_dict import LazyDict
from apispec.utils import OpenAPIVersion
from .common import resolve_schema_cls

##### marshmallow #####

MARSHMALLOW_VERSION_INFO = tuple(
    [int(part) for part in marshmallow.__version__.split('.') if part.isdigit()],
)

# marshmallow field => (JSON Schema type, format)
DEFAULT_FIELD_MAPPING = {
    marshmallow.fields.Integer: ('integer', 'int32'),
    marshmallow.fields.Number: ('number', None),
    marshmallow.fields.Float: ('number', 'float'),
    marshmallow.fields.Decimal: ('number', None),
    marshmallow.fields.String: ('string', None),
    marshmallow.fields.Boolean: ('boolean', None),
    marshmallow.fields.UUID: ('string', 'uuid'),
    marshmallow.fields.DateTime: ('string', 'date-time'),
    marshmallow.fields.Date: ('string', 'date'),
    marshmallow.fields.Time: ('string', None),
    marshmallow.fields.Email: ('string', 'email'),
    marshmallow.fields.URL: ('string', 'url'),
    marshmallow.fields.Dict: ('object', None),
    # Assume base Field and Raw are strings
    marshmallow.fields.Field: ('string', None),
    marshmallow.fields.Raw: ('string', None),
    marshmallow.fields.List: ('array', None),
}


__location_map__ = {
    'query': 'query',
    'querystring': 'query',
    'json': 'body',
    'headers': 'header',
    'cookies': 'cookie',
    'form': 'formData',
    'files': 'formData',
}


# Properties that may be defined in a field's metadata that will be added to the output
# of field2property
# https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#schemaObject
_VALID_PROPERTIES = {
    'format',
    'title',
    'description',
    'default',
    'multipleOf',
    'maximum',
    'exclusiveMaximum',
    'minimum',
    'exclusiveMinimum',
    'maxLength',
    'minLength',
    'pattern',
    'maxItems',
    'minItems',
    'uniqueItems',
    'maxProperties',
    'minProperties',
    'required',
    'enum',
    'type',
    'items',
    'allOf',
    'properties',
    'additionalProperties',
    'readOnly',
    'xml',
    'externalDocs',
    'example',
}

_VALID_PREFIX = 'x-'


class OrderedLazyDict(LazyDict, OrderedDict):
    pass


class OpenAPIConverter(object):
    """Converter generating OpenAPI specification from Marshmallow schemas and fields

    :param str|OpenAPIVersion openapi_version: The OpenAPI version to use.
        Should be in the form '2.x' or '3.x.x' to comply with the OpenAPI standard.
    """
    def __init__(self, openapi_version):
        self.openapi_version = OpenAPIVersion(openapi_version)
        # Schema references
        self.refs = {}
        # Field mappings
        self.field_mapping = DEFAULT_FIELD_MAPPING

    @staticmethod
    def _observed_name(field, name):
        """Adjust field name to reflect `dump_to` and `load_from` attributes.

        :param Field field: A marshmallow field.
        :param str name: Field name
        :rtype: str
        """
        if MARSHMALLOW_VERSION_INFO[0] < 3:
            # use getattr in case we're running against older versions of marshmallow.
            dump_to = getattr(field, 'dump_to', None)
            load_from = getattr(field, 'load_from', None)
            return dump_to or load_from or name
        return field.data_key or name

    def map_to_openapi_type(self, *args):
        """Decorator to set mapping for custom fields.

        ``*args`` can be:

        - a pair of the form ``(type, format)``
        - a core marshmallow field type (in which case we reuse that type's mapping)
        """
        if len(args) == 1 and args[0] in self.field_mapping:
            openapi_type_field = self.field_mapping[args[0]]
        elif len(args) == 2:
            openapi_type_field = args
        else:
            raise TypeError('Pass core marshmallow field type or (type, fmt) pair')

        def inner(field_type):
            self.field_mapping[field_type] = openapi_type_field
            return field_type

        return inner

    def field2choices(self, field, **kwargs):
        """Return the dictionary of OpenAPI field attributes for valid choices definition

        :param Field field: A marshmallow field.
        :rtype: dict
        """
        attributes = {}

        comparable = [
            validator.comparable for validator in field.validators
            if hasattr(validator, 'comparable')
        ]
        if comparable:
            attributes['enum'] = comparable
        else:
            choices = [
                OrderedSet(validator.choices) for validator in field.validators
                if hasattr(validator, 'choices')
            ]
            if choices:
                attributes['enum'] = list(functools.reduce(operator.and_, choices))

        return attributes

    def field2read_only(self, field, **kwargs):
        """Return the dictionary of OpenAPI field attributes for a dump_only field.

        :param Field field: A marshmallow field.
        :rtype: dict
        """
        attributes = {}
        if field.dump_only:
            attributes['readOnly'] = True
        return attributes

    def field2write_only(self, field, **kwargs):
        """Return the dictionary of OpenAPI field attributes for a load_only field.

        :param Field field: A marshmallow field.
        :rtype: dict
        """
        attributes = {}
        if field.load_only and self.openapi_version.major >= 3:
            attributes['writeOnly'] = True
        return attributes

    def field2nullable(self, field, **kwargs):
        """Return the dictionary of OpenAPI field attributes for a nullable field.

        :param Field field: A marshmallow field.
        :rtype: dict
        """
        attributes = {}
        if field.allow_none:
            attributes['x-nullable' if self.openapi_version.major < 3 else 'nullable'] = True
        return attributes

    def field2range(self, field, **kwargs):
        """Return the dictionary of OpenAPI field attributes for a set of
        :class:`Range <marshmallow.validators.Range>` validators.

        :param Field field: A marshmallow field.
        :rtype: dict
        """
        validators = [
            validator for validator in field.validators
            if (
                hasattr(validator, 'min') and
                hasattr(validator, 'max') and
                not hasattr(validator, 'equal')
            )
        ]

        attributes = {}
        for validator in validators:
            if validator.min is not None:
                if hasattr(attributes, 'minimum'):
                    attributes['minimum'] = max(
                        attributes['minimum'],
                        validator.min,
                    )
                else:
                    attributes['minimum'] = validator.min
            if validator.max is not None:
                if hasattr(attributes, 'maximum'):
                    attributes['maximum'] = min(
                        attributes['maximum'],
                        validator.max,
                    )
                else:
                    attributes['maximum'] = validator.max
        return attributes

    def field2length(self, field, **kwargs):
        """Return the dictionary of OpenAPI field attributes for a set of
        :class:`Length <marshmallow.validators.Length>` validators.

        :param Field field: A marshmallow field.
        :rtype: dict
        """
        attributes = {}

        validators = [
            validator for validator in field.validators
            if (
                hasattr(validator, 'min') and
                hasattr(validator, 'max') and
                hasattr(validator, 'equal')
            )
        ]

        is_array = isinstance(
            field, (
                marshmallow.fields.Nested,
                marshmallow.fields.List,
            ),
        )
        min_attr = 'minItems' if is_array else 'minLength'
        max_attr = 'maxItems' if is_array else 'maxLength'

        for validator in validators:
            if validator.min is not None:
                if hasattr(attributes, min_attr):
                    attributes[min_attr] = max(
                        attributes[min_attr],
                        validator.min,
                    )
                else:
                    attributes[min_attr] = validator.min
            if validator.max is not None:
                if hasattr(attributes, max_attr):
                    attributes[max_attr] = min(
                        attributes[max_attr],
                        validator.max,
                    )
                else:
                    attributes[max_attr] = validator.max

        for validator in validators:
            if validator.equal is not None:
                attributes[min_attr] = validator.equal
                attributes[max_attr] = validator.equal
        return attributes

    def field2property(self, field, use_refs=True, dump=True, name=None):
        """Return the JSON Schema property definition given a marshmallow
        :class:`Field <marshmallow.fields.Field>`.

        Will include field metadata that are valid properties of OpenAPI schema objects
        (e.g. "description", "enum", "example").

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#schemaObject

        :param Field field: A marshmallow field.
        :param bool use_refs: Use JSONSchema ``refs``.
        :param bool dump: Introspect dump logic.
        :param str name: The definition name, if applicable, used to construct the $ref value.
        :rtype: dict, a Property Object
        """

        type_, fmt = self.field_mapping.get(type(field), ('string', None))

        ret = {
            'type': type_,
        }

        if fmt:
            ret['format'] = fmt

        default = field.missing
        if default is not marshmallow.missing:
            if callable(default):
                ret['default'] = default()
            else:
                ret['default'] = default

        for attr_func in (
                self.field2choices,
                self.field2read_only,
                self.field2write_only,
                self.field2nullable,
                self.field2range,
                self.field2length,
        ):
            ret.update(attr_func(field))

        if isinstance(field, marshmallow.fields.Nested):
            del ret['type']
            # marshmallow>=2.7.0 compat
            field.metadata.pop('many', None)

            is_unbound_self_referencing = not getattr(field, 'parent', None) and field.nested == 'self'
            if (use_refs and 'ref' in field.metadata) or is_unbound_self_referencing:
                if 'ref' in field.metadata:
                    ref_name = field.metadata['ref']
                else:
                    if not name:
                        raise ValueError(
                            'Must pass `name` argument for self-referencing Nested fields.',
                        )
                    # We need to use the `name` argument when the field is self-referencing and
                    # unbound (doesn't have `parent` set) because we can't access field.schema
                    ref_path = self.get_ref_path()
                    ref_name = '#/{ref_path}/{name}'.format(ref_path=ref_path, name=name)
                ref_schema = {'$ref': ref_name}
                if field.many:
                    ret['type'] = 'array'
                    ret['items'] = ref_schema
                else:
                    if ret:
                        ret.update({'allOf': [ref_schema]})
                    else:
                        ret.update(ref_schema)
            else:
                schema_dict = self.resolve_schema_dict(field.schema, dump=dump)
                if ret and '$ref' in schema_dict:
                    ret.update({'allOf': [schema_dict]})
                else:
                    ret.update(schema_dict)
        elif isinstance(field, marshmallow.fields.List):
            ret['items'] = self.field2property(
                field.container, use_refs=use_refs, dump=dump,
            )
        elif isinstance(field, marshmallow.fields.Dict):
            if MARSHMALLOW_VERSION_INFO[0] >= 3:
                if field.value_container:
                    ret['additionalProperties'] = self.field2property(
                        field.value_container, use_refs=use_refs, dump=dump,
                    )

        # Dasherize metadata that starts with x_
        metadata = {
            key.replace('_', '-') if key.startswith('x_') else key: value
            for key, value in iteritems(field.metadata)
        }
        for key, value in iteritems(metadata):
            if key in _VALID_PROPERTIES or key.startswith(_VALID_PREFIX):
                ret[key] = value
        # Avoid validation error with "Additional properties not allowed"
        # Property "ref" is not valid in this context
        ret.pop('ref', None)

        return ret


    def schema2parameters(self, schema, **kwargs):
        """Return an array of OpenAPI parameters given a given marshmallow
        :class:`Schema <marshmallow.Schema>`. If `default_in` is "body", then return an array
        of a single parameter; else return an array of a parameter for each included field in
        the :class:`Schema <marshmallow.Schema>`.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject
        """
        if hasattr(schema, 'fields'):
            fields = schema.fields
        elif hasattr(schema, '_declared_fields'):
            fields = copy.deepcopy(schema._declared_fields)
        else:
            raise ValueError(
                "{0!r} doesn't have either `fields` or `_declared_fields`".format(schema),
            )
        return self.fields2parameters(fields, schema, **kwargs)


    def fields2parameters(
        self, fields, schema=None, use_refs=True,
        default_in='body', name='body', required=False,
        use_instances=False, description=None, **kwargs
    ):
        """Return an array of OpenAPI parameters given a mapping between field names and
        :class:`Field <marshmallow.Field>` objects. If `default_in` is "body", then return an array
        of a single parameter; else return an array of a parameter for each included field in
        the :class:`Schema <marshmallow.Schema>`.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject

        In OpenAPI3, only "query", "header", "path" or "cookie" are allowed for the location
        of parameters. In OpenAPI 3, "requestBody" is used when fields are in the body.

        This function always returns a list, with a parameter
        for each included field in the :class:`Schema <marshmallow.Schema>`.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#parameterObject
        """
        openapi_default_in = __location_map__.get(default_in, default_in)
        if self.openapi_version.major < 3 and openapi_default_in == 'body':
            if schema is not None:
                prop = self.resolve_schema_dict(schema, dump=False, use_instances=use_instances)
            else:
                prop = self.fields2jsonschema(fields, use_refs=use_refs, dump=False)

            param = {
                'in': openapi_default_in,
                'required': required,
                'name': name,
                'schema': prop,
            }

            if description:
                param['description'] = description

            return [param]

        assert not getattr(schema, 'many', False), \
            "Schemas with many=True are only supported for 'json' location (aka 'in: body')"

        exclude_fields = getattr(getattr(schema, 'Meta', None), 'exclude', [])
        dump_only_fields = getattr(getattr(schema, 'Meta', None), 'dump_only', [])

        parameters = []
        body_param = None
        for field_name, field_obj in iteritems(fields):
            if (field_name in exclude_fields or field_obj.dump_only or field_name in dump_only_fields):
                continue
            param = self.field2parameter(
                field_obj,
                name=self._observed_name(field_obj, field_name),
                use_refs=use_refs,
                default_in=default_in,
            )
            if self.openapi_version.major < 3 and param['in'] == 'body' and body_param is not None:
                body_param['schema']['properties'].update(param['schema']['properties'])
                required_fields = param['schema'].get('required', [])
                if required_fields:
                    body_param['schema'].setdefault('required', []).extend(required_fields)
            else:
                if self.openapi_version.major < 3 and param['in'] == 'body':
                    body_param = param
                parameters.append(param)
        return parameters


    def field2parameter(self, field, name='body', use_refs=True, default_in='body'):
        """Return an OpenAPI parameter as a `dict`, given a marshmallow
        :class:`Field <marshmallow.Field>`.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject
        """
        location = field.metadata.get('location', None)
        prop = self.field2property(field, use_refs=use_refs, dump=False)
        return self.property2parameter(
            prop,
            name=name,
            required=field.required,
            multiple=isinstance(field, marshmallow.fields.List),
            location=location,
            default_in=default_in,
        )


    def property2parameter(
        self, prop, name='body', required=False, multiple=False,
        location=None, default_in='body',
    ):
        """Return the Parameter Object definition for a JSON Schema property.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject

        :param dict prop: JSON Schema property
        :param str name: Field name
        :param bool required: Parameter is required
        :param bool multiple: Parameter is repeated
        :param str location: Location to look for ``name``
        :param str default_in: Default location to look for ``name``
        :raise: TranslationError if arg object cannot be translated to a Parameter Object schema.
        :rtype: dict, a Parameter Object
        """
        openapi_default_in = __location_map__.get(default_in, default_in)
        openapi_location = __location_map__.get(location, openapi_default_in)
        ret = {
            'in': openapi_location,
            'name': name,
        }

        if openapi_location == 'body':
            ret['required'] = False
            ret['name'] = 'body'
            ret['schema'] = {
                'type': 'object',
                'properties': {name: prop} if name else {},
            }
            if name and required:
                ret['schema']['required'] = [name]
        else:
            ret['required'] = required
            if self.openapi_version.major < 3:
                if multiple:
                    ret['collectionFormat'] = 'multi'
                ret.update(prop)
            else:
                if multiple:
                    ret['explode'] = True
                    ret['style'] = 'form'
                if prop.get('description', None):
                    ret['description'] = prop.pop('description')
                ret['schema'] = prop
        return ret

    def schema2jsonschema(self, schema, use_refs=True, dump=True, name=None):
        if hasattr(schema, 'fields'):
            fields = schema.fields
        elif hasattr(schema, '_declared_fields'):
            fields = copy.deepcopy(schema._declared_fields)
        else:
            raise ValueError(
                "{0!r} doesn't have either `fields` or `_declared_fields`".format(schema),
            )

        return self.fields2jsonschema(
            fields, schema, use_refs=use_refs, dump=dump, name=name,
        )


    def fields2jsonschema(self, fields, schema=None, use_refs=True, dump=True, name=None):
        """Return the JSON Schema Object for a given marshmallow
        :class:`Schema <marshmallow.Schema>`. Schema may optionally provide the ``title`` and
        ``description`` class Meta options.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#schemaObject

        Example: ::

            class UserSchema(Schema):
                _id = fields.Int()
                email = fields.Email(description='email address of the user')
                name = fields.Str()

                class Meta:
                    title = 'User'
                    description = 'A registered user'

            OpenAPI.schema2jsonschema(UserSchema)
            # {
            #     'title': 'User', 'description': 'A registered user',
            #     'properties': {
            #         'name': {'required': False,
            #                 'description': '',
            #                 'type': 'string'},
            #         '_id': {'format': 'int32',
            #                 'required': False,
            #                 'description': '',
            #                 'type': 'integer'},
            #         'email': {'format': 'email',
            #                 'required': False,
            #                 'description': 'email address of the user',
            #                 'type': 'string'}
            #     }
            # }

        :param Schema schema: A marshmallow Schema instance or a class object
        :rtype: dict, a JSON Schema Object
        """
        Meta = getattr(schema, 'Meta', None)
        if getattr(Meta, 'fields', None) or getattr(Meta, 'additional', None):
            declared_fields = set(schema._declared_fields.keys())
            if (
                set(getattr(Meta, 'fields', set())) > declared_fields or
                set(getattr(Meta, 'additional', set())) > declared_fields
            ):
                warnings.warn(
                    'Only explicitly-declared fields will be included in the Schema Object. '
                    'Fields defined in Meta.fields or Meta.additional are ignored.',
                )

        jsonschema = {
            'type': 'object',
            'properties': OrderedLazyDict() if getattr(Meta, 'ordered', None) else LazyDict(),
        }

        exclude = set(getattr(Meta, 'exclude', []))

        for field_name, field_obj in iteritems(fields):
            if field_name in exclude or (field_obj.dump_only and not dump):
                continue

            observed_field_name = self._observed_name(field_obj, field_name)
            prop_func = lambda field_obj=field_obj: self.field2property(  # flake8: noqa
                field_obj, use_refs=use_refs, dump=dump, name=name,
            )
            jsonschema['properties'][observed_field_name] = prop_func

            partial = getattr(schema, 'partial', None)
            if field_obj.required:
                if not partial or (is_collection(partial) and field_name not in partial):
                    jsonschema.setdefault('required', []).append(observed_field_name)

        if 'required' in jsonschema:
            jsonschema['required'].sort()

        if Meta is not None:
            if hasattr(Meta, 'title'):
                jsonschema['title'] = Meta.title
            if hasattr(Meta, 'description'):
                jsonschema['description'] = Meta.description

        if getattr(schema, 'many', False):
            jsonschema = {
                'type': 'array',
                'items': jsonschema,
            }

        return jsonschema


    def get_ref_path(self):
        """Return the path for references based on the openapi version

        :param int openapi_version.major: The major version of the OpenAPI standard
            to use. Supported values are 2 and 3.
        """
        ref_paths = {
            2: 'definitions',
            3: 'components/schemas',
        }
        return ref_paths[self.openapi_version.major]

    def resolve_schema_dict(self, schema, dump=True, use_instances=False):
        if isinstance(schema, dict):
            if schema.get('type') == 'array' and 'items' in schema:
                schema['items'] = self.resolve_schema_dict(
                    schema['items'], use_instances=use_instances,
                )
            if schema.get('type') == 'object' and 'properties' in schema:
                schema['properties'] = {
                    k: self.resolve_schema_dict(v, dump=dump, use_instances=use_instances)
                    for k, v in schema['properties'].items()
                }
            return schema
        if isinstance(schema, marshmallow.Schema) and use_instances:
            schema_cls = schema
        else:
            schema_cls = resolve_schema_cls(schema)

        if schema_cls in self.refs:
            ref_path = self.get_ref_path()
            ref_schema = {'$ref': '#/{0}/{1}'.format(ref_path, self.refs[schema_cls])}
            if getattr(schema, 'many', False):
                return {
                    'type': 'array',
                    'items': ref_schema,
                }
            return ref_schema
        if not isinstance(schema, marshmallow.Schema):
            schema = schema_cls
        return self.schema2jsonschema(schema, dump=dump)



class LazyDict(dict):
    """Simple implementation of a lazy dictionary. If a value set is callable
    without parameters, the value will be called the first time is requested
    and stored for future requests.
    E.g:
        dict = LazyDict()
        dict['foo'] = lambda: return 123
        dict['bar'] = dict
        dict['chu'] = dict() # you can also pass non-callable values
        assert dict['foo'] == 123
        assert dict['bar'] == {}
        assert dict['chu'] == {}

    For Internal use only."""

    def values(self):
        return [self.__getitem__(key) for key in self.keys()]

    def items(self):
        return [(key, self.__getitem__(key)) for key in self.keys()]

    def iteritems(self):
        return [(key, self.__getitem__(key)) for key in self.keys()]

    def itervalues(self):
        return [self.__getitem__(key) for key in self.keys()]

    def get(self, k, d=None):
        return self.__getitem__(k) if k in self else d

    def __getitem__(self, key):
        item = super(LazyDict, self).__getitem__(key)
        try:
            self[key] = item = item()
        except TypeError:
            pass
        return item

    def __repr__(self):
        # make dictionaries evaluate
        self.values()
        return super(LazyDict, self).__repr__()

    def __eq__(self, other):
        # make dictionaries evaluate
        self.values()
        if isinstance(other, LazyDict):
            other.values()

        return super(LazyDict, self).__eq__(other)



# -*- coding: utf-8 -*-
"""Flask plugin. Includes a path helper that allows you to pass a view
function to `add_path`. Inspects URL rules and view docstrings.

Passing a view function::

    from flask import Flask

    app = Flask(__name__)

    @app.route('/gists/<gist_id>')
    def gist_detail(gist_id):
        '''Gist detail view.
        ---
        x-extension: metadata
        get:
            responses:
                200:
                    schema:
                        $ref: '#/definitions/Gist'
        '''
        return 'detail for gist {}'.format(gist_id)

    with app.test_request_context():
        spec.add_path(view=gist_detail)
    print(spec.to_dict()['paths'])
    # {'/gists/{gist_id}': {'get': {'responses': {200: {'schema': {'$ref': '#/definitions/Gist'}}}},
    #                  'x-extension': 'metadata'}}

Passing a method view function::

    from flask import Flask
    from flask.views import MethodView

    app = Flask(__name__)

    class GistApi(MethodView):
        '''Gist API.
        ---
        x-extension: metadata
        '''
        def get(self):
           '''Gist view
           ---
           responses:
               200:
                   schema:
                       $ref: '#/definitions/Gist'
           '''
           pass

        def post(self):
           pass

    method_view = GistApi.as_view('gists')
    app.add_url_rule("/gists", view_func=method_view)
    with app.test_request_context():
        spec.add_path(view=method_view)
    print(spec.to_dict()['paths'])
    # {'/gists': {'get': {'responses': {200: {'schema': {'$ref': '#/definitions/Gist'}}}},
    #             'post': {},
    #             'x-extension': 'metadata'}}


"""
from __future__ import absolute_import
import re

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

from flask import current_app
from flask.views import MethodView

from apispec.compat import iteritems
from apispec import Path, BasePlugin, utils
from apispec.exceptions import APISpecError


# from flask-restplus
RE_URL = re.compile(r'<(?:[^:<>]+:)?([^<>]+)>')


class FlaskPlugin(BasePlugin):
    """APISpec plugin for Flask"""

    @staticmethod
    def flaskpath2openapi(path):
        """Convert a Flask URL rule to an OpenAPI-compliant path.

        :param str path: Flask path template.
        """
        return RE_URL.sub(r'{\1}', path)

    @staticmethod
    def _rule_for_view(view):
        view_funcs = current_app.view_functions
        endpoint = None
        for ept, view_func in iteritems(view_funcs):
            if view_func == view:
                endpoint = ept
        if not endpoint:
            raise APISpecError('Could not find endpoint for view {0}'.format(view))

        # WARNING: Assume 1 rule per view function for now
        rule = current_app.url_map._rules_by_endpoint[endpoint][0]
        return rule

    def path_helper(self, view, **kwargs):
        """Path helper that allows passing a Flask view function."""
        rule = self._rule_for_view(view)
        path = self.flaskpath2openapi(rule.rule)
        app_root = current_app.config['APPLICATION_ROOT'] or '/'
        path = urljoin(app_root.rstrip('/') + '/', path.lstrip('/'))
        operations = utils.load_operations_from_docstring(view.__doc__)
        path = Path(path=path, operations=operations)
        if hasattr(view, 'view_class') and issubclass(view.view_class, MethodView):
            operations = {}
            for method in view.methods:
                if method in rule.methods:
                    method_name = method.lower()
                    method = getattr(view.view_class, method_name)
                    docstring_yaml = utils.load_yaml_from_docstring(method.__doc__)
                    operations[method_name] = docstring_yaml or dict()
            path.operations.update(operations)
        return path


# Deprecated interface
def setup(spec):
    """Setup for the plugin.

    .. deprecated:: 0.39.0
        Use FlaskPlugin class.
    """
    plugin = FlaskPlugin()
    plugin.init_spec(spec)
    spec.plugins.append(plugin)



# -*- coding: utf-8 -*-
"""Contains the main `APISpec` class.
"""
from .core import APISpec, Path
from .plugin import BasePlugin

__version__ = '0.39.0'
__author__ = 'Steven Loria, Jérôme Lafréchoux, and contributors'
__license__ = 'MIT'


__all__ = [
    'APISpec',
    'Path',
    'BasePlugin',
]



# -*- coding: utf-8 -*-
"""Base class for Plugin classes."""


from .exceptions import PluginMethodNotImplementedError

class BasePlugin(object):
    """Base class for APISpec plugin classes."""
    def init_spec(self, spec):
        """Initialize plugin with APISpec object

        :param APISpec spec: APISpec object this plugin instance is attached to
        """

    def definition_helper(self, name, definition, **kwargs):
        """Must return definition as a dict."""
        raise PluginMethodNotImplementedError

    def path_helper(self, path=None, operations=None, **kwargs):
        """Should return a Path instance. Any other return value type is ignored"""
        raise PluginMethodNotImplementedError

    def operation_helper(self, path=None, operations=None, **kwargs):
        """Should mutate operations. Return value ignored."""
        raise PluginMethodNotImplementedError

    def response_helper(self, method, status_code, **kwargs):
        """Should return a dict to update the response description.

        Returning None is equivalent to returning an empty dictionary.
        """
        raise PluginMethodNotImplementedError



# -*- coding: utf-8 -*-
# flake8: noqa
import sys

PY2 = int(sys.version[0]) == 2

if PY2:
    text_type = unicode
    binary_type = str
    string_types = (str, unicode)
    unicode = unicode
    basestring = basestring
    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()
else:
    text_type = str
    binary_type = bytes
    string_types = (str,)
    unicode = str
    basestring = (str, bytes)
    iterkeys = lambda d: d.keys()
    itervalues = lambda d: d.values()
    iteritems = lambda d: d.items()



# -*- coding: utf-8 -*-
"""Core apispec classes and functions."""
import re
import warnings
from collections import OrderedDict
import copy

import yaml

from apispec.compat import iterkeys, iteritems, PY2, unicode
from apispec.lazy_dict import LazyDict
from .exceptions import PluginError, APISpecError, PluginMethodNotImplementedError
from .utils import OpenAPIVersion

VALID_METHODS = [
    'get',
    'post',
    'put',
    'patch',
    'delete',
    'head',
    'options',
]


def clean_operations(operations, openapi_major_version):
    """Ensure that all parameters with "in" equal to "path" are also required
    as required by the OpenAPI specification, as well as normalizing any
    references to global parameters.

    See https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#parameterObject.

    :param dict operations: Dict mapping status codes to operations
    :param int openapi_major_version: The major version of the OpenAPI standard
        to use. Supported values are 2 and 3.
    """
    def get_ref(param, openapi_major_version):
        if isinstance(param, dict):
            return param

        ref_paths = {
            2: 'parameters',
            3: 'components/parameters',
        }
        ref_path = ref_paths[openapi_major_version]
        return {'$ref': '#/{0}/{1}'.format(ref_path, param)}

    for operation in (operations or {}).values():
        if 'parameters' in operation:
            parameters = operation.get('parameters')
            for parameter in parameters:
                if (
                    isinstance(parameter, dict) and
                    'in' in parameter and parameter['in'] == 'path'
                ):
                    parameter['required'] = True
            operation['parameters'] = [
                get_ref(p, openapi_major_version)
                for p in parameters
            ]


class Path(object):
    """Represents an OpenAPI Path object.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#pathsObject

    :param str path: The path template, e.g. ``"/pet/{petId}"``
    :param str method: The HTTP method.
    :param dict operation: The operation object, as a `dict`. See
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#operationObject
    :param str|OpenAPIVersion openapi_version: The OpenAPI version to use.
        Should be in the form '2.x' or '3.x.x' to comply with the OpenAPI standard.
    """
    def __init__(self, path=None, operations=None, openapi_version='2.0'):
        self.path = path
        operations = operations or OrderedDict()
        openapi_version = OpenAPIVersion(openapi_version)
        clean_operations(operations, openapi_version.major)
        invalid = {key for key in
                   set(iterkeys(operations)) - set(VALID_METHODS)
                   if not key.startswith('x-')}
        if invalid:
            raise APISpecError(
                'One or more HTTP methods are invalid: {0}'.format(', '.join(invalid)),
            )
        self.operations = operations

    def to_dict(self):
        if not self.path:
            raise APISpecError('Path template is not specified')
        return {
            self.path: self.operations,
        }

    def update(self, path):
        if path.path:
            self.path = path.path
        self.operations.update(path.operations)


class APISpec(object):
    """Stores metadata that describes a RESTful API using the OpenAPI specification.

    :param str title: API title
    :param str version: API version
    :param tuple plugins: Import paths to plugins.
    :param dict info: Optional dict to add to `info`
        See https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#infoObject
    :param callable schema_name_resolver: Callable to generate the schema definition name.
        This parameter is deprecated. It is now a parameter of MarshmallowPlugin.
    :param str|OpenAPIVersion openapi_version: The OpenAPI version to use.
        Should be in the form '2.x' or '3.x.x' to comply with the OpenAPI standard.
    :param dict options: Optional top-level keys
        See https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#swagger-object
    """
    def __init__(
        self, title, version, plugins=(), info=None,
        schema_name_resolver=None, openapi_version='2.0', **options
    ):
        self.info = {
            'title': title,
            'version': version,
        }
        self.info.update(info or {})

        self.openapi_version = OpenAPIVersion(openapi_version)

        self.options = options
        if schema_name_resolver is not None:
            warnings.warn(
                'schema_name_resolver parameter is deprecated. '
                'It is now a parameter of MarshmallowPlugin.',
                DeprecationWarning,
            )
        self.schema_name_resolver = schema_name_resolver
        # Metadata
        self._definitions = {}
        self._parameters = {}
        self._tags = []
        self._paths = OrderedDict()

        # Plugins
        plugins = list(plugins)  # Cast to list in case a generator is passed
        # Backward compatibility: plugins can be passed as objects or strings
        self.plugins = list(p for p in plugins if not isinstance(p, str))
        for plugin in self.plugins:
            plugin.init_spec(self)

        # Deprecated interface
        # Plugins and helpers
        self.old_plugins = {}
        self._definition_helpers = []
        self._path_helpers = []
        self._operation_helpers = []
        # {'get': {200: [my_helper]}}
        self._response_helpers = {}
        old_plugins = list(p for p in plugins if isinstance(p, str))
        for plugin_path in old_plugins:
            self.setup_plugin(plugin_path)

    def to_dict(self):
        ret = {
            'info': self.info,
            'paths': self._paths,
            'tags': self._tags,
        }

        if self.openapi_version.major == 2:
            ret['swagger'] = self.openapi_version.vstring
            ret['definitions'] = self._definitions
            ret['parameters'] = self._parameters
            ret.update(self.options)

        elif self.openapi_version.major == 3:
            ret['openapi'] = self.openapi_version.vstring
            options = copy.deepcopy(self.options)
            components = options.pop('components', {})

            # deep update components object
            definitions = components.pop('schemas', {})
            definitions.update(self._definitions)
            parameters = components.pop('parameters', {})
            parameters.update(self._parameters)

            ret['components'] = dict(
                schemas=definitions,
                parameters=parameters,
                **components
            )
            ret.update(options)

        return ret

    def to_yaml(self):
        return yaml.dump(self.to_dict(), Dumper=YAMLDumper)

    def add_parameter(self, param_id, location, **kwargs):
        """ Add a parameter which can be referenced.

        :param str param_id: identifier by which parameter may be referenced.
        :param str location: location of the parameter.
        :param dict kwargs: parameter fields.
        """
        if 'name' not in kwargs:
            kwargs['name'] = param_id
        kwargs['in'] = location
        self._parameters[param_id] = kwargs

    def add_tag(self, tag):
        """ Store information about a tag.

        :param dict tag: the dictionary storing information about the tag.
        """
        self._tags.append(tag)

    def add_path(self, path=None, operations=None, **kwargs):
        """Add a new path object to the spec.

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#pathsObject

        :param str|Path|None path: URL Path component or Path instance
        :param dict|None operations: describes the http methods and options for `path`
        :param dict kwargs: parameters used by any path helpers see :meth:`register_path_helper`
        """
        def normalize_path(path):
            if path and 'basePath' in self.options:
                pattern = '^{0}'.format(re.escape(self.options['basePath']))
                path = re.sub(pattern, '', path)
            return path

        if isinstance(path, Path):
            path.path = normalize_path(path.path)
            if operations:
                path.operations.update(operations)
        else:
            path = Path(
                path=normalize_path(path),
                operations=operations,
                openapi_version=self.openapi_version,
            )

        # Execute path helpers
        for plugin in self.plugins:
            try:
                ret = plugin.path_helper(path=path, operations=path.operations, **kwargs)
            except PluginMethodNotImplementedError:
                continue
            if isinstance(ret, Path):
                ret.path = normalize_path(ret.path)
                path.update(ret)
        # Deprecated interface
        for func in self._path_helpers:
            try:
                ret = func(
                    self, path=path, operations=path.operations, **kwargs
                )
            except TypeError:
                continue
            if isinstance(ret, Path):
                ret.path = normalize_path(ret.path)
                path.update(ret)
        if not path.path:
            raise APISpecError('Path template is not specified')

        # Execute operation helpers
        for plugin in self.plugins:
            try:
                plugin.operation_helper(path=path, operations=path.operations, **kwargs)
            except PluginMethodNotImplementedError:
                continue
        # Deprecated interface
        for func in self._operation_helpers:
            func(self, path=path, operations=path.operations, **kwargs)

        # Execute response helpers
        # TODO: cache response helpers output for each (method, status_code) couple
        for method, operation in iteritems(path.operations):
            if method in VALID_METHODS and 'responses' in operation:
                for status_code, response in iteritems(operation['responses']):
                    for plugin in self.plugins:
                        try:
                            response.update(plugin.response_helper(method, status_code, **kwargs) or {})
                        except PluginMethodNotImplementedError:
                            continue
        # Deprecated interface
        # Rule is that method + http status exist in both operations and helpers
        methods = set(iterkeys(path.operations)) & set(iterkeys(self._response_helpers))
        for method in methods:
            responses = path.operations[method]['responses']
            statuses = set(iterkeys(responses)) & set(iterkeys(self._response_helpers[method]))
            for status_code in statuses:
                for func in self._response_helpers[method][status_code]:
                    responses[status_code].update(
                        func(self, **kwargs),
                    )

        self._paths.setdefault(path.path, path.operations).update(path.operations)

    def definition(
        self, name, properties=None, enum=None, description=None, extra_fields=None,
        **kwargs
    ):
        """Add a new definition to the spec.

        .. note::

            If you are using `apispec.ext.marshmallow`, you can pass fields' metadata as
            additional keyword arguments.

            For example, to add ``enum`` to your field: ::

                status = fields.String(required=True, enum=['open', 'closed'])

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#definitionsObject
        """
        ret = {}
        # Execute all helpers from plugins
        for plugin in self.plugins:
            try:
                ret.update(plugin.definition_helper(name, definition=ret, **kwargs))
            except PluginMethodNotImplementedError:
                continue
        # Deprecated interface
        for func in self._definition_helpers:
            try:
                ret.update(func(self, name, definition=ret, **kwargs))
            except TypeError:
                continue
        if properties:
            ret['properties'] = properties
        if enum:
            ret['enum'] = enum
        if description:
            ret['description'] = description
        if extra_fields:
            ret.update(extra_fields)
        self._definitions[name] = ret

    # Deprecated PLUGIN INTERFACE

    # adapted from Sphinx
    def setup_plugin(self, path):
        """Import and setup a plugin. No-op if called twice
        for the same plugin.

        :param str path: Import path to the plugin.
        :raise: PluginError if the given plugin is invalid.
        """
        warnings.warn(
            'Old style plugins are deprecated. Use classes instead. '
            'See https://apispec.readthedocs.io/en/latest/writing_plugins.html.',
            DeprecationWarning,
        )
        if path in self.old_plugins:
            return
        try:
            mod = __import__(
                path, globals=None, locals=None, fromlist=('setup', ),
            )
        except ImportError as err:
            raise PluginError(
                'Could not import plugin "{0}"\n\n{1}'.format(path, err),
            )
        if not hasattr(mod, 'setup'):
            raise PluginError('Plugin "{0}" has no setup(spec) function'.format(path))
        else:
            # Each plugin gets a dict to store arbitrary data
            self.old_plugins[path] = {}
            mod.setup(self)

    # Deprecated helpers interface

    def register_definition_helper(self, func):
        """Register a new definition helper. The helper **must** meet the following conditions:

        - Receive the `APISpec` instance as the first argument.
        - Receive the definition `name` as the second argument.
        - Include ``**kwargs`` in its signature.
        - Return a `dict` representation of the definition's Schema object.

        The helper may define any named arguments after the `name` argument.
        ``kwargs`` will include (among other things):
        - definition (dict): current state of the definition

        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#definitionsObject

        :param callable func: The definition helper function.
        """
        warnings.warn(
            'Helper functions are deprecated. Use plugin classes. '
            'See https://apispec.readthedocs.io/en/latest/writing_plugins.html.',
            DeprecationWarning,
        )
        self._definition_helpers.append(func)

    def register_path_helper(self, func):
        """Register a new path helper. The helper **must** meet the following conditions:

        - Receive the `APISpec` instance as the first argument.
        - Include ``**kwargs`` in signature.
        - Return a `apispec.core.Path` object.

        The helper may define any named arguments in its signature.
        """
        warnings.warn(
            'Helper functions are deprecated. Use plugin classes. '
            'See https://apispec.readthedocs.io/en/latest/writing_plugins.html.',
            DeprecationWarning,
        )
        self._path_helpers.append(func)

    def register_operation_helper(self, func):
        """Register a new operation helper. The helper **must** meet the following conditions:

        - Receive the `APISpec` instance as the first argument.
        - Receive ``operations`` as a keyword argument.
        - Include ``**kwargs`` in signature.

        The helper may define any named arguments in its signature.
        """
        warnings.warn(
            'Helper functions are deprecated. Use plugin classes. '
            'See https://apispec.readthedocs.io/en/latest/writing_plugins.html.',
            DeprecationWarning,
        )
        self._operation_helpers.append(func)

    def register_response_helper(self, func, method, status_code):
        """Register a new response helper. The helper **must** meet the following conditions:

        - Receive the `APISpec` instance as the first argument.
        - Include ``**kwargs`` in signature.
        - Return a `dict` response object.

        The helper may define any named arguments in its signature.
        """
        warnings.warn(
            'Helper functions are deprecated. Use plugin classes. '
            'See https://apispec.readthedocs.io/en/latest/writing_plugins.html.',
            DeprecationWarning,
        )
        method = method.lower()
        if method not in self._response_helpers:
            self._response_helpers[method] = {}
        self._response_helpers[method].setdefault(status_code, []).append(func)


class YAMLDumper(yaml.Dumper):

    @staticmethod
    def _represent_dict(dumper, instance):
        return dumper.represent_mapping('tag:yaml.org,2002:map', instance.items())

    if PY2:
        @staticmethod
        def _represent_unicode(_, uni):
            return yaml.ScalarNode(tag=u'tag:yaml.org,2002:str', value=uni)


if PY2:
    yaml.add_representer(unicode, YAMLDumper._represent_unicode, Dumper=YAMLDumper)
yaml.add_representer(OrderedDict, YAMLDumper._represent_dict, Dumper=YAMLDumper)
yaml.add_representer(LazyDict, YAMLDumper._represent_dict, Dumper=YAMLDumper)
yaml.add_representer(Path, YAMLDumper._represent_dict, Dumper=YAMLDumper)



# -*- coding: utf-8 -*-
"""Bottle plugin. Includes a path helper that allows you to pass a view function
to `add_path`.
::

    from bottle import route, default_app
    app = default_app()
    @route('/gists/<gist_id>')
    def gist_detail(gist_id):
        '''Gist detail view.
        ---
        get:
            responses:
                200:
                    schema:
                        $ref: '#/definitions/Gist'
        '''
        return 'detail for gist {}'.format(gist_id)

    spec.add_path(view=gist_detail)
    print(spec.to_dict()['paths'])
    # {'/gists/{gist_id}': {'get': {'responses': {200: {'schema': {'$ref': '#/definitions/Gist'}}}}}}
"""
from __future__ import absolute_import
import re

from bottle import default_app

from apispec import Path, BasePlugin, utils
from apispec.exceptions import APISpecError


RE_URL = re.compile(r'<(?:[^:<>]+:)?([^<>]+)>')

_default_app = default_app()


class BottlePlugin(BasePlugin):
    """APISpec plugin for Bottle"""

    @staticmethod
    def bottle_path_to_openapi(path):
        return RE_URL.sub(r'{\1}', path)

    @staticmethod
    def _route_for_view(app, view):
        endpoint = None
        for route in app.routes:
            if route._context['callback'] == view:
                endpoint = route
                break
        if not endpoint:
            raise APISpecError('Could not find endpoint for route {0}'.format(view))
        return endpoint

    def path_helper(self, view, operations, **kwargs):
        """Path helper that allows passing a bottle view function."""
        operations = utils.load_operations_from_docstring(view.__doc__)
        app = kwargs.get('app', _default_app)
        route = self._route_for_view(app, view)
        bottle_path = self.bottle_path_to_openapi(route.rule)
        return Path(path=bottle_path, operations=operations)


# Deprecated interface
def setup(spec):
    """Setup for the plugin.

    .. deprecated:: 0.39.0
        Use BottlePlugin class.
    """
    plugin = BottlePlugin()
    plugin.init_spec(spec)
    spec.plugins.append(plugin)



# -*- coding: utf-8 -*-
"""Marshmallow plugin for apispec. Allows passing a marshmallow
`Schema` to `APISpec.definition <apispec.APISpec.definition>`
and `APISpec.add_path <apispec.APISpec.add_path>` (for responses).

Requires marshmallow>=2.0.

::

    from pprint import pprint

    from marshmallow import Schema, fields

    class UserSchema(Schema):
        id = fields.Int(dump_only=True)
        name = fields.Str(description="The user's name")

    spec.definition('User', schema=UserSchema)
    pprint(spec.to_dict()['definitions'])
    # {'User': {'properties': {'id': {'format': 'int32', 'type': 'integer'},
    #                         'name': {'description': "The user's name",
    #                                 'type': 'string'}},
    #         'type': 'object'}}

"""
from __future__ import absolute_import

import marshmallow

from apispec import Path, BasePlugin, utils
from .common import resolve_schema_cls, resolve_schema_instance
from .openapi import OpenAPIConverter


class MarshmallowPlugin(BasePlugin):
    """APISpec plugin handling marshmallow schemas

    :param callable schema_name_resolver: Callable to generate the schema definition name.
        Receives the `Schema` class and returns the name to be used in refs within
        the generated spec.

        Example: ::

            def schema_name_resolver(schema):
                return schema.__name__
    """
    def __init__(self, schema_name_resolver=None):
        super(MarshmallowPlugin, self).__init__()
        self.schema_name_resolver = schema_name_resolver
        self.spec = None
        self.openapi_version = None
        self.openapi = None

    def init_spec(self, spec):
        super(MarshmallowPlugin, self).init_spec(spec)
        self.spec = spec
        self.openapi_version = spec.openapi_version
        self.openapi = OpenAPIConverter(openapi_version=spec.openapi_version)

    def inspect_schema_for_auto_referencing(self, original_schema_instance):
        """Parse given schema instance and reference eventual nested schemas
        :param original_schema_instance: schema to parse
        """
        # schema_name_resolver must be provided to use this function
        assert self.schema_name_resolver

        for field in original_schema_instance.fields.values():
            nested_schema_class = None

            if isinstance(field, marshmallow.fields.Nested):
                nested_schema_class = resolve_schema_cls(field.schema)

            elif isinstance(field, marshmallow.fields.List) \
                    and isinstance(field.container, marshmallow.fields.Nested):
                nested_schema_class = resolve_schema_cls(field.container.schema)

            if nested_schema_class and nested_schema_class not in self.openapi.refs:
                definition_name = self.schema_name_resolver(
                    nested_schema_class,
                )
                if definition_name:
                    self.spec.definition(
                        definition_name,
                        schema=nested_schema_class,
                    )

    def resolve_parameters(self, parameters):
        resolved = []
        for parameter in parameters:
            if not isinstance(parameter.get('schema', {}), dict):
                schema_cls = resolve_schema_cls(parameter['schema'])
                if issubclass(schema_cls, marshmallow.Schema) and 'in' in parameter:
                    del parameter['schema']
                    resolved += self.openapi.schema2parameters(
                        schema_cls,
                        default_in=parameter.pop('in'), **parameter
                    )
                    continue
            self.resolve_schema(parameter)
            resolved.append(parameter)
        return resolved

    def resolve_schema_in_request_body(self, request_body):
        """Function to resolve a schema in a requestBody object - modifies then
        response dict to convert Marshmallow Schema object or class into dict
        """
        content = request_body['content']
        for content_type in content:
            schema = content[content_type]['schema']
            content[content_type]['schema'] = self.openapi.resolve_schema_dict(schema)

    def resolve_schema(self, data):
        """Function to resolve a schema in a parameter or response - modifies the
        corresponding dict to convert Marshmallow Schema object or class into dict

        :param APISpec spec: `APISpec` containing refs.
        :param dict data: the parameter or response dictionary that may contain a schema
        """
        if self.openapi_version.major < 3:
            if 'schema' in data:
                data['schema'] = self.openapi.resolve_schema_dict(data['schema'])
        else:
            if 'content' in data:
                for content_type in data['content']:
                    schema = data['content'][content_type]['schema']
                    data['content'][content_type]['schema'] = self.openapi.resolve_schema_dict(schema)

    def map_to_openapi_type(self, *args):
        """Decorator to set mapping for custom fields.

        ``*args`` can be:

        - a pair of the form ``(type, format)``
        - a core marshmallow field type (in which case we reuse that type's mapping)

        Examples: ::

            @ma_plugin.map_to_openapi_type('string', 'uuid')
            class MyCustomField(Integer):
                # ...

            @ma_plugin.map_to_openapi_type(Integer)  # will map to ('integer', 'int32')
            class MyCustomFieldThatsKindaLikeAnInteger(Integer):
                # ...
        """
        return self.openapi.map_to_openapi_type(*args)

    def definition_helper(self, name, schema, **kwargs):
        """Definition helper that allows using a marshmallow
        :class:`Schema <marshmallow.Schema>` to provide OpenAPI
        metadata.

        :param type|Schema schema: A marshmallow Schema class or instance.
        """

        schema_cls = resolve_schema_cls(schema)
        schema_instance = resolve_schema_instance(schema)

        # Store registered refs, keyed by Schema class
        self.openapi.refs[schema_cls] = name

        json_schema = self.openapi.schema2jsonschema(schema_instance, name=name)

        # Auto reference schema if schema_name_resolver
        if self.schema_name_resolver:
            self.inspect_schema_for_auto_referencing(schema_instance)

        return json_schema

    def path_helper(self, view=None, **kwargs):
        """Path helper that allows passing a Schema as a response. Responses can be
        defined in a view's docstring.
        ::

            from pprint import pprint

            from my_app import Users, UserSchema

            class UserHandler:
                def get(self, user_id):
                    '''Get a user endpoint.
                    ---
                    description: Get a user
                    responses:
                        200:
                            description: A user
                            schema: UserSchema
                    '''
                    user = Users.get(id=user_id)
                    schema = UserSchema()
                    return schema.dumps(user)

            urlspec = (r'/users/{user_id}', UserHandler)
            spec.add_path(urlspec=urlspec)
            pprint(spec.to_dict()['paths'])
            # {'/users/{user_id}': {'get': {'description': 'Get a user',
            #                               'responses': {200: {'description': 'A user',
            #                                                   'schema': {'$ref': '#/definitions/User'}}}}}}

        ::

            from pprint import pprint

            from my_app import Users, UserSchema

            class UsersHandler:
                def get(self):
                    '''Get users endpoint.
                    ---
                    description: Get a list of users
                    responses:
                        200:
                            description: A list of user
                            schema:
                                type: array
                                items: UserSchema
                    '''
                    users = Users.all()
                    schema = UserSchema(many=True)
                    return schema.dumps(users)

            urlspec = (r'/users', UsersHandler)
            spec.add_path(urlspec=urlspec)
            pprint(spec.to_dict()['paths'])
            # {'/users': {'get': {'description': 'Get a list of users',
            #                     'responses': {200: {'description': 'A list of users',
            #                                         'schema': {'type': 'array',
            #                                                    'items': {'$ref': '#/definitions/User'}}}}}}}

        """
        operations = (
            kwargs.get('operations') or
            (view and utils.load_operations_from_docstring(view.__doc__))
        )
        if not operations:
            return None
        operations = operations.copy()
        return Path(operations=operations)

    def operation_helper(self, operations, **kwargs):
        for operation in operations.values():
            if not isinstance(operation, dict):
                continue
            if 'parameters' in operation:
                operation['parameters'] = self.resolve_parameters(operation['parameters'])
            if self.openapi_version.major >= 3:
                if 'requestBody' in operation:
                    self.resolve_schema_in_request_body(operation['requestBody'])
            for response in operation.get('responses', {}).values():
                self.resolve_schema(response)


# Deprecated interface
def setup(spec):
    """Setup for the plugin.

    .. deprecated:: 0.39.0
        Use MarshmallowPlugin class.
    """
    plugin = MarshmallowPlugin(schema_name_resolver=spec.schema_name_resolver)
    plugin.init_spec(spec)
    spec.plugins.append(plugin)



# -*- coding: utf-8 -*-
"""Tornado plugin. Includes a path helper that allows you to pass an urlspec (path-handler pair)
object to `add_path`.
::

    from pprint import pprint

    from tornado.web import RequestHandler

    class HelloHandler(RequestHandler):
        def get(self):
            '''Get a greeting endpoint.
            ---
            description: Get a greeting
            responses:
                200:
                    description: A greeting to the client
                    schema:
                        $ref: '#/definitions/Greeting'
            '''
            self.write("hello")

    urlspec = (r'/hello', HelloHandler)
    spec.add_path(urlspec=urlspec)
    pprint(spec.to_dict()['paths'])
    # {'/hello': {'get': {'description': 'Get a greeting',
    #                     'responses': {200: {'description': 'A greeting to the '
    #                                                     'client',
    #                                         'schema': {'$ref': '#/definitions/Greeting'}}}}}}

"""
from __future__ import absolute_import
import inspect
import sys
from tornado.web import URLSpec

from apispec import Path, BasePlugin, utils
from apispec.exceptions import APISpecError


class TornadoPlugin(BasePlugin):
    """APISpec plugin for Tornado"""

    @staticmethod
    def _operations_from_methods(handler_class):
        """Generator of operations described in handler's http methods

        :param handler_class:
        :type handler_class: RequestHandler descendant
        """
        for httpmethod in utils.PATH_KEYS:
            method = getattr(handler_class, httpmethod)
            operation_data = utils.load_yaml_from_docstring(method.__doc__)
            if operation_data:
                operation = {httpmethod: operation_data}
                yield operation

    @staticmethod
    def tornadopath2openapi(urlspec, method):
        """Convert Tornado URLSpec to OpenAPI-compliant path.

        :param urlspec:
        :type urlspec: URLSpec
        :param method: Handler http method
        :type method: function
        """
        if sys.version_info >= (3, 3):
            args = list(inspect.signature(method).parameters.keys())[1:]
        else:
            if getattr(method, '__tornado_coroutine__', False):
                method = method.__wrapped__
            args = inspect.getargspec(method).args[1:]
        params = tuple('{{{}}}'.format(arg) for arg in args)
        try:
            path_tpl = urlspec.matcher._path
        except AttributeError:  # tornado<4.5
            path_tpl = urlspec._path
        path = (path_tpl % params)
        if path.count('/') > 1:
            path = path.rstrip('/?*')
        return path

    @staticmethod
    def _extensions_from_handler(handler_class):
        """Returns extensions dict from handler docstring

        :param handler_class:
        :type handler_class: RequestHandler descendant
        """
        extensions = utils.load_yaml_from_docstring(handler_class.__doc__) or {}
        return extensions

    def path_helper(self, urlspec, operations, **kwargs):
        """Path helper that allows passing a Tornado URLSpec or tuple."""
        if not isinstance(urlspec, URLSpec):
            urlspec = URLSpec(*urlspec)
        if not operations:
            operations = {}
            for operation in self._operations_from_methods(urlspec.handler_class):
                operations.update(operation)
        if not operations:
            raise APISpecError(
                'Could not find endpoint for urlspec {0}'.format(urlspec),
            )
        params_method = getattr(urlspec.handler_class, list(operations.keys())[0])
        path = self.tornadopath2openapi(urlspec, params_method)
        extensions = self._extensions_from_handler(urlspec.handler_class)
        operations.update(extensions)
        return Path(path=path, operations=operations)


# Deprecated interface
def setup(spec):
    """Setup for the plugin.

    .. deprecated:: 0.39.0
        Use TornadoPlugin class.
    """
    plugin = TornadoPlugin()
    plugin.init_spec(spec)
    spec.plugins.append(plugin)



# -*- coding: utf-8 -*-
"""Various utilities for parsing OpenAPI operations from docstrings and validating against
the OpenAPI spec.
"""
import re
import json
import warnings

from distutils import version

import yaml

from apispec.compat import iteritems
from apispec import exceptions

# from django.contrib.admindocs.utils
def trim_docstring(docstring):
    """Uniformly trims leading/trailing whitespace from docstrings.

    Based on http://www.python.org/peps/pep-0257.html#handling-docstring-indentation
    """
    if not docstring or not docstring.strip():
        return ''
    # Convert tabs to spaces and split into lines
    lines = docstring.expandtabs().splitlines()
    indent = min(len(line) - len(line.lstrip()) for line in lines if line.lstrip())
    trimmed = [lines[0].lstrip()] + [line[indent:].rstrip() for line in lines[1:]]
    return '\n'.join(trimmed).strip()

# from rest_framework.utils.formatting
def dedent(content):
    """
    Remove leading indent from a block of text.
    Used when generating descriptions from docstrings.
    Note that python's `textwrap.dedent` doesn't quite cut it,
    as it fails to dedent multiline docstrings that include
    unindented text on the initial line.
    """
    whitespace_counts = [len(line) - len(line.lstrip(' '))
                         for line in content.splitlines()[1:] if line.lstrip()]

    # unindent the content if needed
    if whitespace_counts:
        whitespace_pattern = '^' + (' ' * min(whitespace_counts))
        content = re.sub(re.compile(whitespace_pattern, re.MULTILINE), '', content)

    return content.strip()

def load_yaml_from_docstring(docstring):
    """Loads YAML from docstring."""
    split_lines = trim_docstring(docstring).split('\n')

    # Cut YAML from rest of docstring
    for index, line in enumerate(split_lines):
        line = line.strip()
        if line.startswith('---'):
            cut_from = index
            break
    else:
        return None

    yaml_string = '\n'.join(split_lines[cut_from:])
    yaml_string = dedent(yaml_string)
    return yaml.load(yaml_string)


PATH_KEYS = set([
    'get',
    'put',
    'post',
    'delete',
    'options',
    'head',
    'patch',
])

def load_operations_from_docstring(docstring):
    """Return a dictionary of OpenAPI operations parsed from a
    a docstring.
    """
    doc_data = load_yaml_from_docstring(docstring)
    if doc_data:
        return {
            key: val for key, val in iteritems(doc_data)
            if key in PATH_KEYS or key.startswith('x-')
        }
    else:
        return None

def validate_spec(spec):
    """Validate the output of an :class:`APISpec` object against the
    OpenAPI specification.

    Note: Requires installing apispec with the ``[validation]`` extras.
    ::

        pip install 'apispec[validation]'

    :raise: apispec.exceptions.OpenAPIError if validation fails.
    """
    try:
        import prance
    except ImportError as error:  # re-raise with a more verbose message
        exc_class = type(error)
        raise exc_class(
            'validate_spec requires prance to be installed. '
            'You can install all validation requirements using:\n'
            "    pip install 'apispec[validation]'",
        )
    parser_kwargs = {}
    if spec.openapi_version.version[0] == 3:
        parser_kwargs['backend'] = 'openapi-spec-validator'
    try:
        prance.BaseParser(spec_string=json.dumps(spec.to_dict()), **parser_kwargs)
    except prance.ValidationError as err:
        raise exceptions.OpenAPIError(*err.args)
    else:
        return True

def validate_swagger(spec):
    """
    .. deprecated:: 0.38.0
        Use `apispec.utils.validate_spec` instead.
    """
    warnings.warn(
        'apispec.utils.validate_swagger is deprecated. Use apispec.utils.validate_spec instead.',
        DeprecationWarning,
    )
    return validate_spec(spec)


class OpenAPIVersion(version.LooseVersion, object):
    """OpenAPI version

    :param str|OpenAPIVersion openapi_version: OpenAPI version

    Parses an OpenAPI version expressed as string. Provides shortcut to digits
    (major, minor, patch).

        Example: ::

            ver = OpenAPIVersion('3.0.1')
            assert ver.major == 3
            assert ver.minor == 0
            assert ver.patch == 1
            assert ver.vstring == '3.0.1'
            assert str(ver) == '3.0.1'
    """
    MIN_INCLUSIVE_VERSION = version.LooseVersion('2.0')
    MAX_EXCLUSIVE_VERSION = version.LooseVersion('4.0')

    def __init__(self, openapi_version):
        if isinstance(openapi_version, version.LooseVersion):
            openapi_version = openapi_version.vstring
        if not self.MIN_INCLUSIVE_VERSION <= openapi_version < self.MAX_EXCLUSIVE_VERSION:
            raise exceptions.APISpecError(
                'Not a valid OpenAPI version number: {}'.format(openapi_version),
            )
        super(OpenAPIVersion, self).__init__(openapi_version)

    @property
    def major(self):
        return self.version[0]

    @property
    def minor(self):
        return self.version[1]

    @property
    def patch(self):
        return self.version[2]



# -*- coding: utf-8 -*-
"""Utilities to get schema instances/classes"""

import marshmallow


def resolve_schema_instance(schema):
    """Return schema instance for given schema (instance or class)

    :param type|Schema|str schema: instance, class or class name of marshmallow.Schema
    :return: schema instance of given schema (instance or class)
    """
    if isinstance(schema, type) and issubclass(schema, marshmallow.Schema):
        return schema()
    if isinstance(schema, marshmallow.Schema):
        return schema
    return marshmallow.class_registry.get_class(schema)()


def resolve_schema_cls(schema):
    """Return schema class for given schema (instance or class)

    :param type|Schema|str: instance, class or class name of marshmallow.Schema
    :return: schema class of given schema (instance or class)
    """
    if isinstance(schema, type) and issubclass(schema, marshmallow.Schema):
        return schema
    if isinstance(schema, marshmallow.Schema):
        return type(schema)
    return marshmallow.class_registry.get_class(schema)



# -*- coding: utf-8 -*-
"""Exception classes."""
import warnings

class APISpecError(Exception):
    """Base class for all apispec-related errors."""

class PluginError(APISpecError):
    """Raised when a plugin cannot be found or is invalid."""

class PluginMethodNotImplementedError(APISpecError, NotImplementedError):
    """Raised when calling an unimplemented helper method in a plugin"""

class OpenAPIError(APISpecError):
    """Raised when a OpenAPI spec validation fails."""

class SwaggerError(OpenAPIError):
    """
    .. deprecated:: 0.38.0
        Use `apispec.exceptions.OpenAPIError` instead.
    """
    def __init__(self, *args, **kwargs):
        warnings.warn(
            'SwaggerError is deprecated. Use OpenAPIError instead.',
            DeprecationWarning,
        )
        super(SwaggerError, self).__init__(*args, **kwargs)
